@@ -26,13 +26,17 @@ pub const DC_OVER_DOT_3: char = '';
 /// Three dots above a tengwa, with one below the others.
 pub const DC_OVER_DOT_3_INV: char = '';
 
+/// A line above a tengwa, which curls to the left.
 pub const DC_OVER_HOOK_L_1: char = '';
 pub const DC_OVER_HOOK_L_2: char = '';
 
+/// A line above a tengwa, which curls to the right.
 pub const DC_OVER_HOOK_R_1: char = '';
 pub const DC_OVER_HOOK_R_2: char = '';
 
+/// A horizontal overline.
 pub const DC_OVER_LINE: char = '';
+/// A wavy horizontal overline.
 pub const DC_OVER_WAVE: char = '';
 
 
@@ -48,23 +52,35 @@ pub const DC_UNDER_DOT_2: char = '';
 /// Three dots below a tengwa, with one below the others.
 pub const DC_UNDER_DOT_3: char = '';
 
+/// A line below a tengwa, which curls to the left.
 pub const DC_UNDER_HOOK_L_1: char = '';
+/// A line below a tengwa, which curls to the right.
 pub const DC_UNDER_HOOK_R_1: char = '';
+/// A horizontal underline.
 pub const DC_UNDER_LINE_H: char = '';
+/// A vertical line below a tengwa. Has various meanings but usually pertains to
+///     the vowel marking above it.
 pub const DC_UNDER_LINE_V: char = '';
 
+/// An unfilled circle below a tengwa. Most often used to denote the least
+///     significant digit in a duodecimal figure.
 pub const DC_UNDER_RING: char = '';
 
 
+/// Marking to denote a sound that leads into a "w" sound.
 pub const MOD_LABIAL: char = DC_OVER_WAVE;
 /// Long/double consonant.
 pub const MOD_LONG_CONS: char = DC_UNDER_LINE_H;
 /// Long vowel after consonant.
 pub const MOD_LONG_VOWEL: char = DC_UNDER_LINE_V;
 
+/// Marking to denote a sound preceded by M or N.
 pub const MOD_NASAL: char = DC_OVER_LINE;
+/// Marking to denote a sound that leads into a "y" sound.
 pub const MOD_PALATAL: char = DC_UNDER_DOT_2;
+/// An ornate flourish added to a tengwa to indicate a following sibilant.
 pub const MOD_SARINCE_L: char = '';
+/// A small hook added to a tengwa to indicate a following sibilant.
 pub const MOD_SARINCE_R: char = '';
 
 /// Index in this array corresponds to the numerical value of the character.
@@ -75,32 +91,59 @@ pub const NUMERAL: [char; 12] = [
 ];
 
 
-pub const PUNCT_DOT_1: &str = "";
-pub const PUNCT_DOT_2: &str = "";
-pub const PUNCT_DOT_3: &str = "";
-pub const PUNCT_DOT_4: &str = "";
-pub const PUNCT_DOT_5: &str = "";
+/// A single dot positioned inside the preceding character.
+pub const PUNCT_DOT_0: char = '';
+pub const PUNCT_DOT_1: char = '';
+pub const PUNCT_DOT_2: char = '';
+pub const PUNCT_DOT_3: char = '';
+pub const PUNCT_DOT_4: char = '';
+pub const PUNCT_DOT_5: char = '';
 
-pub const PUNCT_DOT_S1: &str = "  ";
-pub const PUNCT_DOT_TRI: &str = "";
-pub const PUNCT_DOT_DIAM: &str = "";
+pub const PUNCT_EXCLAM: char = '';
+pub const PUNCT_THORIN: char = '';
+pub const PUNCT_INTERR: char = '';
 
-pub const PUNCT_EXCLAM: &str = "";
-pub const PUNCT_INTERR: &str = "";
+pub const PUNCT_LINE_1: char = '';
+pub const PUNCT_LINE_2: char = '';
 
-pub const PUNCT_LINE_1: &str = "";
-pub const PUNCT_LINE_2: &str = "";
-pub const PUNCT_LINE_S1: &str = "  ";
-pub const PUNCT_LINE_S2: &str = "  ";
-
-pub const PUNCT_PAREN: &str = "";
-pub const PUNCT_PAREN_L: &str = "";
-pub const PUNCT_PAREN_R: &str = "";
+pub const PUNCT_PAREN: char = '';
+pub const PUNCT_PAREN_L: char = '';
+pub const PUNCT_PAREN_R: char = '';
 
+/// A compound marking most often denoting the end of a passage.
 pub const PUNCT_EOF: &str = "";
 
 
-pub const TEHTA_CIRCUMFLEX: Tehta = Tehta::basic(DC_OVER_CIRCUMFLEX);
+/// Convert non-tengwar punctuation marker into one from the tengwar block.
+///     Where unambiguous replacements are not known, this is chosen, admittedly
+///     arbitrarily, based on superficial similarity.
+///
+/// Only single characters are returned; compound punctuation thus may be
+///     constructed from similar basic marks, such as `:-` and `::`.
+pub const fn punctuation(chr: char) -> Option<char> {
+    Some(match chr {
+        '\'' | '.' | ',' | '·' => PUNCT_DOT_1,
+        ':' | ';' => PUNCT_DOT_2,
+        '⁝' | '︙' => PUNCT_DOT_3,
+        '⁘' | '⁛' | '…' => PUNCT_DOT_4,
+        '⸭' => PUNCT_DOT_5,
+
+        '-' => PUNCT_LINE_1,
+        '=' => PUNCT_LINE_2,
+
+        '?' => PUNCT_INTERR,
+        '!' => PUNCT_EXCLAM,
+        '|' | '‖' => PUNCT_PAREN,
+        '(' | '[' | '“' => PUNCT_PAREN_L,
+        ')' | ']' | '”' | '„' => PUNCT_PAREN_R,
+
+        _ => { return None; }
+    })
+}
+
+
+/// A diacritical marking resembling a circumflex.
+pub const TEHTA_CIRCUMFLEX: Tehta = Tehta::Single(DC_OVER_CIRCUMFLEX);
 
 
 /// The diacritic for an `A` vowel, in its standard three-dot form.
@@ -117,12 +160,12 @@ const _A: char = DC_OVER_CIRCUMFLEX;
 mod _vowels {
     use super::*;
 
-    pub const TEHTA_A: Tehta = Tehta::basic(_A);
-    pub const TEHTA_E: Tehta = Tehta::basic(DC_OVER_ACUTE_1);
-    pub const TEHTA_I: Tehta = Tehta::basic(DC_OVER_DOT_1);
-    pub const TEHTA_O: Tehta = Tehta::basic(DC_OVER_HOOK_R_1);
-    pub const TEHTA_U: Tehta = Tehta::basic(DC_OVER_HOOK_L_1);
-    pub const TEHTA_Y: Tehta = Tehta::basic(DC_OVER_DOT_2);
+    pub const TEHTA_A: Tehta = Tehta::Single(_A);
+    pub const TEHTA_E: Tehta = Tehta::Single(DC_OVER_ACUTE_1);
+    pub const TEHTA_I: Tehta = Tehta::Single(DC_OVER_DOT_1);
+    pub const TEHTA_O: Tehta = Tehta::Single(DC_OVER_HOOK_R_1);
+    pub const TEHTA_U: Tehta = Tehta::Single(DC_OVER_HOOK_L_1);
+    pub const TEHTA_Y: Tehta = Tehta::Single(DC_OVER_DOT_2);
 }
 
 
@@ -130,13 +173,13 @@ mod _vowels {
 mod _vowels {
     use super::*;
 
-    pub const TEHTA_A: Tehta = Tehta::basic(_A);
-    pub const TEHTA_E: Tehta = Tehta::with_double(DC_OVER_ACUTE_1);
-    // pub const TEHTA_I: Tehta = Tehta::with_double(DC_OVER_DOT_1);
-    pub const TEHTA_I: Tehta = Tehta::basic(DC_OVER_DOT_1);
-    pub const TEHTA_O: Tehta = Tehta::with_double(DC_OVER_HOOK_R_1);
-    pub const TEHTA_U: Tehta = Tehta::with_double(DC_OVER_HOOK_L_1);
-    pub const TEHTA_Y: Tehta = Tehta::basic(DC_OVER_DOT_2);
+    pub const TEHTA_A: Tehta = Tehta::Single(_A);
+    pub const TEHTA_E: Tehta = Tehta::Double(DC_OVER_ACUTE_1);
+    // pub const TEHTA_I: Tehta = Tehta::Double(DC_OVER_DOT_1);
+    pub const TEHTA_I: Tehta = Tehta::Single(DC_OVER_DOT_1);
+    pub const TEHTA_O: Tehta = Tehta::Double(DC_OVER_HOOK_R_1);
+    pub const TEHTA_U: Tehta = Tehta::Double(DC_OVER_HOOK_L_1);
+    pub const TEHTA_Y: Tehta = Tehta::Single(DC_OVER_DOT_2);
 }
 
 
@@ -144,19 +187,20 @@ mod _vowels {
 mod _vowels {
     use super::*;
 
-    pub const TEHTA_A: Tehta = Tehta::basic(_A);
-    pub const TEHTA_E: Tehta = Tehta::with_variant(DC_OVER_ACUTE_1, DC_OVER_ACUTE_2);
-    // pub const TEHTA_I: Tehta = Tehta::with_variant(DC_OVER_DOT_1, DC_OVER_DOT_2);
-    pub const TEHTA_I: Tehta = Tehta::basic(DC_OVER_DOT_1);
-    pub const TEHTA_O: Tehta = Tehta::with_variant(DC_OVER_HOOK_R_1, DC_OVER_HOOK_R_2);
-    pub const TEHTA_U: Tehta = Tehta::with_variant(DC_OVER_HOOK_L_1, DC_OVER_HOOK_L_2);
-    pub const TEHTA_Y: Tehta = Tehta::basic(DC_OVER_DOT_2);
+    pub const TEHTA_A: Tehta = Tehta::Single(_A);
+    pub const TEHTA_E: Tehta = Tehta::Altern(DC_OVER_ACUTE_1, DC_OVER_ACUTE_2);
+    // pub const TEHTA_I: Tehta = Tehta::Altern(DC_OVER_DOT_1, DC_OVER_DOT_2);
+    pub const TEHTA_I: Tehta = Tehta::Single(DC_OVER_DOT_1);
+    pub const TEHTA_O: Tehta = Tehta::Altern(DC_OVER_HOOK_R_1, DC_OVER_HOOK_R_2);
+    pub const TEHTA_U: Tehta = Tehta::Altern(DC_OVER_HOOK_L_1, DC_OVER_HOOK_L_2);
+    pub const TEHTA_Y: Tehta = Tehta::Single(DC_OVER_DOT_2);
 }
 
 
 pub use _vowels::*;
 
 
+/// The T-series, with an open bow to the right.
 pub const TEMA_TINCO: Tema = Tema {
     single_dn: '', // Tinco
     double_dn: '', // Ando
@@ -167,6 +211,7 @@ pub const TEMA_TINCO: Tema = Tema {
     single_ex: '',
     double_ex: '',
 };
+/// The P-series, with a closed bow to the right.
 pub const TEMA_PARMA: Tema = Tema {
     single_dn: '', // Parma
     double_dn: '', // Umbar
@@ -177,6 +222,7 @@ pub const TEMA_PARMA: Tema = Tema {
     single_ex: '',
     double_ex: '',
 };
+/// The C-series, with an open bow to the left.
 pub const TEMA_CALMA: Tema = Tema {
     single_dn: '', // Calma
     double_dn: '', // Anga
@@ -187,6 +233,7 @@ pub const TEMA_CALMA: Tema = Tema {
     single_ex: '',
     double_ex: '',
 };
+/// The Q-series, with a closed bow to the left.
 pub const TEMA_QESSE: Tema = Tema {
     single_dn: '', // Qessë
     double_dn: '', // Ungwë
@@ -198,8 +245,10 @@ pub const TEMA_QESSE: Tema = Tema {
     double_ex: '',
 };
 
-pub const TENGWA_CURL_SINGLE: char = '';
-pub const TENGWA_CURL_DOUBLE: char = '';
+/// A variant of Vala with an extra hook.
+pub const TENGWA_CURL_SINGLE: char = '';
+/// A variant of Malta with an extra hook.
+pub const TENGWA_CURL_DOUBLE: char = '';
 
 pub const TENGWA_ROMEN: char = '';
 pub const TENGWA_ARDA: char = '';
@@ -223,9 +272,11 @@ pub const TENGWA_HALLA: char = '';
 pub const TENGWA_WAIA: char = '';
 pub const TENGWA_OSSE: char = '';
 
+/// "Zero-Width Joiner", used for forming ligatures.
 pub const ZWJ: char = '‍';
 
 
+/// Return a carrier mark appropriate for whether a vowel is long.
 pub const fn carrier(long: bool) -> char {
     if long { CARRIER_LONG } else { CARRIER_SHORT }
 }
@@ -248,6 +299,7 @@ fn int(mut n: isize, base: isize) -> (bool, Vec<usize>) {
 }
 
 
+/// Render an integer into text form as tengwar.
 pub fn int_10(n: isize) -> String {
     let (neg, digits): (bool, Vec<usize>) = int(n, 10);
     let mut out = String::with_capacity(neg as usize + digits.len() * 6);
@@ -264,6 +316,7 @@ pub fn int_10(n: isize) -> String {
 }
 
 
+/// Render an integer into text form as tengwar, in duodecimal (base-12).
 pub fn int_12(n: isize) -> String {
     let (neg, digits): (bool, Vec<usize>) = int(n, 12);
     let mut out = String::with_capacity(neg as usize + digits.len() * 6);
@@ -286,51 +339,54 @@ pub fn int_12(n: isize) -> String {
 }
 
 
+/// A diacritical vowel marker that may be rendered in an alternate "long" form.
 #[derive(Clone)]
-pub struct Tehta {
-    pub base: char,
-    pub long: Option<char>,
-    pub double: bool,
+pub enum Tehta {
+    /// There is only one form that this diacritic can take. The long form must
+    ///     be rendered with an extended carrier mark.
+    Single(char),
+    /// There is one form that this diacritic can take, but it may be repeated
+    ///     for the long form.
+    Double(char),
+    /// This diacritic has an alternate character to represent its long form.
+    Altern(char, char),
 }
 
 
 impl Tehta {
-    pub const fn basic(base: char) -> Self {
-        Self { base, long: None, double: false }
-    }
-
-    pub const fn with_double(base: char) -> Self {
-        Self { base, long: None, double: true }
-    }
-
-    pub const fn with_variant(base: char, long: char) -> Self {
-        Self { base, long: Some(long), double: false }
-    }
-
+    /// Write this Tehta into a Formatter. The provided boolean argument will
+    ///     determine whether the basic short form or the variable long form is
+    ///     written.
     pub fn write(&self, f: &mut Formatter<'_>, long: bool) -> fmt::Result {
-        if long {
-            if self.double {
-                f.write_char(self.base)?;
-                f.write_char(self.base)
-            } else {
-                match self.long {
-                    Some(variant) => f.write_char(variant),
-                    None => {
-                        // f.write_char(MOD_LONG_VOWEL)?;
-                        f.write_char(carrier(true))?;
-                        f.write_char(self.base)
-                    }
+        match self {
+            &Self::Single(mark) => {
+                if long {
+                    f.write_char(carrier(true))?;
+                }
+
+                f.write_char(mark)
+            }
+            &Self::Double(mark) => {
+                if long {
+                    f.write_char(mark)?;
+                }
+
+                f.write_char(mark)
+            }
+            &Self::Altern(mark, alt) => {
+                if long {
+                    f.write_char(mark)
+                } else {
+                    f.write_char(alt)
                 }
             }
-        } else {
-            f.write_char(self.base)
         }
     }
 
     /// Returns true if the long variant of this Tehta would be written with the
     ///     extended "Ára" Telco.
     pub const fn uses_ara(&self) -> bool {
-        !self.double && self.long.is_none()
+        matches!(self, Self::Single(..))
     }
 }
 
@@ -370,11 +426,13 @@ pub struct Tema {
 }
 
 
+/// Check whether a tengwa has an inverted variant.
 pub const fn can_be_nuquerna(c: char) -> bool {
     c == TENGWA_SILME || c == TENGWA_ESSE
 }
 
 
+/// Convert a tengwa to its inverted variant.
 pub const fn nuquerna(c: char) -> char {
     match c {
         TENGWA_SILME => TENGWA_SILME_NUQ,
@@ -384,6 +442,8 @@ pub const fn nuquerna(c: char) -> char {
 }
 
 
+/// Check whether a base tengwa is suitable for ligation with the extended
+///     carrier mark. This is to some degree based on opinion.
 const fn ligates_with_ara(base: char) -> bool {
     (TEMA_TINCO.single_dn <= base && base <= TENGWA_HWESTA_SINDARINWA)
         && base != TENGWA_SILME_NUQ
@@ -411,24 +471,37 @@ pub const fn ligature_valid(prev: &Glyph, next: &Glyph) -> bool {
 }
 
 
+/// Check whether a base tengwa is suitable to receive a sa-rincë. This is to
+///     some degree based on opinion.
+pub const fn rince_valid(base: char) -> bool {
+    !matches!(base, '' | '' | '' | '' | '' | '')
+}
+
+
 /// Choose the appropriate form of sa-rincë for a base tengwa.
 #[cfg(not(feature = "alt-rince"))]
-pub const fn mod_rince(_base: char) -> char {
-    MOD_SARINCE_L
+pub const fn mod_rince(_base: char, _is_final: bool) -> char {
+    MOD_SARINCE_R
 }
 
 
 /// Choose the appropriate form of sa-rincë for a base tengwa.
 #[cfg(feature = "alt-rince")]
-pub const fn mod_rince(base: char) -> char {
-    match base {
-        '' | '' | '' | '' | '' | '' | '' | ''
-        => MOD_SARINCE_R,
-        _ => MOD_SARINCE_L,
+pub const fn mod_rince(base: char, is_final: bool) -> char {
+    if is_final {
+        match base {
+            ''..='' | '' | '' | '' => MOD_SARINCE_L,
+            _ => MOD_SARINCE_R,
+        }
+    } else {
+        MOD_SARINCE_R
     }
 }
 
 
+/// A single base tengwa, and all of its modifications. This includes the vowel
+///     marking, flags for additional diacritics, flags for consonant and vowel
+///     length, and an indicator of finality.
 #[derive(Clone, Default)]
 pub struct Glyph {
     /// A consonant character.
@@ -451,10 +524,13 @@ pub struct Glyph {
     /// Indicates whether a long vowel using the extended "Ára" Telco should be
     ///     placed before this glyph.
     pub long_first: bool,
+    /// This glyph is the final one in a word, and may use a more ornate rincë.
+    pub is_final: bool,
 }
 
 
 impl Glyph {
+    /// Define a new empty glyph.
     pub const fn new() -> Self {
         Self {
             cons: None,
@@ -466,9 +542,11 @@ impl Glyph {
             long_cons: false,
             long_vowel: false,
             long_first: false,
+            is_final: false,
         }
     }
 
+    /// Define a glyph with both a base character and a diacritical `Tehta`.
     pub const fn new_both(cons: char, vowel: Tehta) -> Self {
         Self {
             cons: Some(cons),
@@ -477,6 +555,7 @@ impl Glyph {
         }
     }
 
+    /// Define a glyph with only a base character. It may be marked as Long.
     pub const fn new_cons(cons: char, long: bool) -> Self {
         Self {
             cons: Some(cons),
@@ -485,6 +564,7 @@ impl Glyph {
         }
     }
 
+    /// Define a glyph with only a diacritical Tehta. It may be marked as Long.
     pub const fn new_vowel(vowel: Tehta, long: bool) -> Self {
         Self {
             vowel: Some(vowel),
@@ -493,21 +573,28 @@ impl Glyph {
         }
     }
 
+    /// Mark this glyph as being labialized. It will be rendered with a wavy
+    ///     overbar.
     pub const fn with_labial(mut self) -> Self {
         self.labial = true;
         self
     }
 
+    /// Mark this glyph as being nasalized. It will be rendered overlined.
     pub const fn with_nasal(mut self) -> Self {
         self.nasal = true;
         self
     }
 
+    /// Mark this glyph as being palatalized. It will be rendered with a pair of
+    ///     dots below it.
     pub const fn with_palatal(mut self) -> Self {
         self.palatal = true;
         self
     }
 
+    /// Mark this glyph as being followed by a sibilant. It may be rendered with
+    ///     a flourish.
     pub const fn with_silme(mut self) -> Self {
         self.silme = true;
         self
@@ -516,11 +603,13 @@ impl Glyph {
 
 
 impl Glyph {
+    /// Determine the base character to be used for this glyph. If one is not
+    ///     set, an appropriate "carrier" mark will be returned instead.
     pub const fn base(&self) -> char {
         match self {
             #[cfg(feature = "nuquernar")]
             &Glyph { cons: Some(con), vowel: Some(ref vowel), long_vowel, .. } if {
-                can_be_nuquerna(con) && (!long_vowel || !vowel.uses_ara())
+                can_be_nuquerna(con) && !(long_vowel && vowel.uses_ara())
             } => nuquerna(con),
 
             &Glyph { cons: Some(con), .. } => con,
@@ -534,6 +623,7 @@ impl Glyph {
             cons, vowel, silme,
             nasal, labial, palatal,
             long_cons, long_vowel, long_first,
+            is_final,
         } = self;
 
         #[cfg_attr(feature = "nuquernar", allow(unused_mut))]
@@ -542,7 +632,7 @@ impl Glyph {
         let vowel_post: Option<&Tehta> = match vowel {
             Some(tehta) => {
                 if long && *long_first {
-                    if tehta.uses_ara() && !ligatures {
+                    if !ligatures && tehta.uses_ara() {
                         tehta.write(f, true)?;
                         None
                     } else if cfg!(not(feature = "nuquernar"))
@@ -610,7 +700,7 @@ impl Glyph {
         }
 
         if *silme {
-            f.write_char(mod_rince(base))?;
+            f.write_char(mod_rince(base, *is_final))?;
         }
 
         Ok(())
